﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.Linq;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Templates;
using Clarion.Framework.Extensions;
using WorldServerLibrary.Model;
using System.Threading;
using System.Diagnostics;

namespace ClarionSimulation
{
    /// <summary>

    /// This event will be fired at the beginning of each cognitive cycle in order to get the current visual sensor information.
    /// </summary>
    /// <returns>The Visual Sensor Information Itens</returns> 
    public delegate IList<Thing> InputVisualSensorialInformationEventHandler();

    /// <summary>
    /// This event will be fired at the end of each cognitive cycle in order to translate the selected action to a World Server Command
    /// </summary>
    /// <param name="externalAction">The external action that was choice</param>
    public delegate String OutputActionChunkEventHandler(ClarionAgentActionType externalAction, Thing whichThing, String message);

    /// <summary>
    /// Public enum that represents all possibilities of agent actions
    /// </summary>
    public enum ClarionAgentActionType
    {
        DO_NOTHING,
        ROTATE_CLOCKWISE,
		EAT_FOOD,
		COLLECT_JEWEL,
		GO_POSITION,
		DELIVER_LEAFLET,
		CONTOUR_OBSTACLE,
		CREATE_THINGS,
		TALK,
		LISTEN
    }
	
	
	public enum jewelColors
	{
		red,
		green,
		blue,
		yellow,
		magenta,
		white,
		NUM_OF_COLORS
	}
	
	
	
	public enum communicationCycles
	{
		OPEN_GREETING,
		MESSAGE,
		CLOSE_GREETING,
		NUM_OF_CYCLES
	}
	
	
	
	
	
    public class ClarionAgent
    {
        #region Constants

		public string[] communicationMessages = {"hello", "carring", "bye"};
		
        /// <summary>
        /// Constant that represents the Visual Sensor
        /// </summary>
        private String SENSOR_VISUAL_DIMENSION = "VisualSensor";

        /// <summary>
        /// Constant that represents that there is at least one wall ahead
        /// </summary>
        private String DIMENSION_WALL_AHEAD = "WallAhead";
		
		private double MAXIMUM_DISTANCE = 1000.0;
		private double COLLECT_DISTANCE = 30.0;
		private double CONTOUR_DISTANCE = 20.0;
		private double DELIVER_DISTANCE = 30.0;
		
		private double TALK_DISTANCE = 60.0;
		
		private double MINIMUM_FUEL = 300.0;
		
		private int CYCLES_TOLERANCE = 5;
		
		private int LISTENING_THRESHOLD     = 1;
		private int WRONG_MESSAGE_THRESHOLD = 1;
		
		private String EAT_FOOD = "eatFood";
		private String FULFILL_LEAFLET = "fulfillLeaflet";

        #endregion

        #region Properties

        #region Simulation

        /// <summary>
        /// If this value is greater than zero, the agent will have a finite number of cognitive cycle. Otherwise, it will have infinite cycles.
        /// </summary>
        public double MaxNumberOfCognitiveCycles = -1;

        /// <summary>
        /// Current cognitive cycle number
        /// </summary>
        private double CurrentCognitiveCycle = 0;

        /// <summary>
        /// Time between cognitive cycle in miliseconds
        /// </summary>
        public Int32 TimeBetweenCognitiveCycles = 0;

        /// <summary>
        /// A thread Class that will handle the simulation process
        /// </summary>
        private Thread runThread;
		
        #endregion

        #region Agent

        /// <summary>
        /// The agent 
        /// </summary>
        private Clarion.Framework.Agent CurrentAgent;

        /// <summary>
        /// For each cognitive cycle, this event will be called in order to the agent receives the current sensorial information
        /// </summary>
        public event InputVisualSensorialInformationEventHandler OnNewVisualSensorialInformation;

        /// <summary>
        /// For each cognitive cycle, this event will be called when the agent selects one action
        /// </summary>
        public event OutputActionChunkEventHandler OnNewExternalActionSelected;
		
		
		private Creature currentCreature;
		
		private GoalChunk eatFood;
		private GoalChunk fulfillLeaflet;
		
		private SimplifiedQBPNetwork movementNet;
		
		private SimplifiedQBPNetwork talkingNet;
		
		
		private Thing deliverySpot = null;
		
		private Leaflet completedLeaflet = null;
		
		
        #endregion

        #region Perception Input

        /// <summary>
        /// Perception input to indicates a wall ahead
        /// </summary>
        private DimensionValuePair InputWallAhead;
		
		private DimensionValuePair leafletCompletion;
		
		private DimensionValuePair pos_x;
		private DimensionValuePair pos_y;
		
		private DimensionValuePair missing_red;
		private DimensionValuePair missing_green;
		private DimensionValuePair missing_blue;
		private DimensionValuePair missing_yellow;
		private DimensionValuePair missing_magenta;
		private DimensionValuePair missing_white;
		
		private DimensionValuePair hasTalkedBefore;
		private DimensionValuePair hasListenedBefore;
		private DimensionValuePair hasListenedNoise;
		private DimensionValuePair hasListenedNothing;
		
		private DimensionValuePair commOpenGreeting;
		private DimensionValuePair commMessage;
		private DimensionValuePair commCloseGreeting;
		
		private communicationCycles currentCycle;
		
		private Boolean justTalked   = false;
		private Boolean justListened = false;
		
		private Boolean justListenedNoise   = false;
		private Boolean justListenedNothing = false;
						
		private Boolean creatureNear      = false;
		private Boolean communicationDone = false;

		private int communicationAttempts = 0;
		private int listeningAttempts     = 0;
		private int listenedWrongMessage  = 0;
		
		

		private List<Thing> jewels;
		private List<Thing> foods;
		
		private Hashtable creatures = null;
		
		private List<Leaflet> leaflets;
		
		private List<Thing> collectedJewels;		
		
		private int[] jewelsNeeded;
		private double[] jewelsActivation;
		
		
		
        #endregion

        #region Action Output

        /// <summary>
        /// Output action that makes the agent to rotate clockwise
        /// </summary>
        private ExternalActionChunk OutputRotateClockwise;

        private ExternalActionChunk outputDeliverLeaflet;
		
		private ExternalActionChunk outputContourObstacle;

        #endregion

        #endregion

        #region Constructor

        public ClarionAgent()
        {
			World.LoggingLevel = TraceLevel.Verbose;
			
            // Initialize the agent
            CurrentAgent = World.NewAgent("Current Agent");

            // Initialize Input Information
            InputWallAhead = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_WALL_AHEAD);
			
			leafletCompletion = World.NewDimensionValuePair ("leaflet", "completion");
			
			pos_x = World.NewDimensionValuePair ("creature", "x");
			pos_y = World.NewDimensionValuePair ("creature", "y");
			
			missing_red     = World.NewDimensionValuePair ("missing", "red");
		    missing_green   = World.NewDimensionValuePair ("missing", "green");
			missing_blue    = World.NewDimensionValuePair ("missing", "blue");
			missing_yellow  = World.NewDimensionValuePair ("missing", "yellow");
			missing_magenta = World.NewDimensionValuePair ("missing", "magenta");
			missing_white   = World.NewDimensionValuePair ("missing", "white");
				

            // Initialize Output actions
            OutputRotateClockwise = World.NewExternalActionChunk (ClarionAgentActionType.ROTATE_CLOCKWISE.ToString());
			OutputRotateClockwise.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.ROTATE_CLOCKWISE.ToString()));

            outputDeliverLeaflet = World.NewExternalActionChunk (ClarionAgentActionType.DELIVER_LEAFLET.ToString());
			outputDeliverLeaflet.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.DELIVER_LEAFLET.ToString()));
			
			outputContourObstacle = World.NewExternalActionChunk (ClarionAgentActionType.CONTOUR_OBSTACLE.ToString());
			outputContourObstacle.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.CONTOUR_OBSTACLE.ToString()));
			
            //Create thread to simulation
            runThread = new Thread (RunThread);
			Console.WriteLine ("Agent started");
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Run the Simulation in World Server 3d Environment
        /// </summary>
        public void Run()
        {                
			Console.WriteLine ("Running ...");
            // Setup Agent to run
            if (runThread != null && !runThread.IsAlive)
            {
                SetupAgentInfraStructure();

                if (OnNewVisualSensorialInformation != null)
                {
                    // Start Simulation Thread                
                    runThread.Start(null);
                }
            }
        }

        /// <summary>
        /// Abort the current Simulation
        /// </summary>
        /// <param name="deleteAgent">If true beyond abort the current simulation it will die the agent.</param>
        public void Abort(Boolean deleteAgent)
        {   Console.WriteLine ("Aborting ...");
            if (runThread != null && runThread.IsAlive)
            {
                runThread.Abort();
            }

            if (CurrentAgent != null && deleteAgent)
            {
                CurrentAgent.Die();
            }
        }

        #endregion

        #region Setup Agent Methods

        /// <summary>
        /// Setup agent infra structure (ACS, NACS, MS and MCS)
        /// </summary>
        private void SetupAgentInfraStructure()
        {	
			jewels = new List<Thing>();
			foods = new List<Thing>();
			
			creatures = new Hashtable();
			
			
			collectedJewels = new List<Thing>();

			jewelsNeeded     = new int[(int) jewelColors.NUM_OF_COLORS];
			jewelsActivation = new double[(int) jewelColors.NUM_OF_COLORS];
			
            // Setup the ACS Subsystem
            SetupACS(); 
			
			SetupTalkingNetwork();
        }
		
		
		
        private void SetupTalkingNetwork()
        {            
            
			hasTalkedBefore      = World.NewDimensionValuePair ("lastOperation", "talked");
			hasListenedBefore    = World.NewDimensionValuePair ("lastOperation", "listened");
				
			hasListenedNoise     = World.NewDimensionValuePair ("lastListening", "noise");
			hasListenedNothing   = World.NewDimensionValuePair ("lastListening", "nothing");
			
			commOpenGreeting     = World.NewDimensionValuePair ("currentCommCycle", "openGreeting");
			commMessage          = World.NewDimensionValuePair ("currentCommCycle", "message");
			commCloseGreeting    = World.NewDimensionValuePair ("currentCommCycle", "closeGreeting");
			
			// Create the message chunks
			
			ExternalActionChunk outputSayHello = World.NewExternalActionChunk ("say_hello");			
			outputSayHello.Add (World.NewDimensionValuePair ("message", communicationMessages[(int) communicationCycles.OPEN_GREETING]));
			outputSayHello.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.TALK.ToString()));

			ExternalActionChunk outputSayGoodbye = World.NewExternalActionChunk ("say_bye");
			outputSayGoodbye.Add (World.NewDimensionValuePair ("message", communicationMessages[(int) communicationCycles.CLOSE_GREETING]));
			outputSayGoodbye.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.TALK.ToString()));

			ExternalActionChunk outputSayMessage = World.NewExternalActionChunk ("say_message");
			outputSayMessage.Add (World.NewDimensionValuePair ("message", communicationMessages[(int) communicationCycles.MESSAGE]));
			outputSayMessage.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.TALK.ToString()));	
			
			// Create multiple listening outputs to compensate the multiple talking
			
			ExternalActionChunk outputListen_1 = World.NewExternalActionChunk ("listen_1");
			outputListen_1.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.LISTEN.ToString()));

			ExternalActionChunk outputListen_2 = World.NewExternalActionChunk ("listen_2");
			outputListen_2.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.LISTEN.ToString()));
			
			ExternalActionChunk outputListen_3 = World.NewExternalActionChunk ("listen_3");
			outputListen_3.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.LISTEN.ToString()));
			
			
			talkingNet = AgentInitializer.InitializeImplicitDecisionNetwork (CurrentAgent, 
			                                                                 SimplifiedQBPNetwork.Factory, 
			                                                                 (EligibilityChecker) CustomEligibilityCheckerCommNetwork);
			
			
			// Still need to add as input the existing creatures
			
			talkingNet.Input.Add (commOpenGreeting);
			talkingNet.Input.Add (commMessage);
			talkingNet.Input.Add (commCloseGreeting);
			
			talkingNet.Input.Add (hasTalkedBefore);
			talkingNet.Input.Add (hasListenedBefore);
			talkingNet.Input.Add (hasListenedNoise);
			talkingNet.Input.Add (hasListenedNothing);
						
			
			// The talking/listening operations are the outputs
			
			talkingNet.Output.Add (outputSayHello);
			talkingNet.Output.Add (outputSayGoodbye);
			talkingNet.Output.Add (outputSayMessage);
			talkingNet.Output.Add (outputListen_1);
			talkingNet.Output.Add (outputListen_2);
			talkingNet.Output.Add (outputListen_3);
			
			talkingNet.Parameters.LEARNING_RATE = 1;
			
			CurrentAgent.Commit (talkingNet);
			
			
			currentCycle = communicationCycles.OPEN_GREETING;
        }

        /// <summary>
        /// Setup the ACS subsystem
        /// </summary>
        private void SetupACS()
        {

            // Create contour obstacle rule

			SupportCalculator contourObstacleSupportCalculator = FixedRuleDelegateToContourObstacle;
			
            FixedRule ruleContourObstacle = AgentInitializer.InitializeActionRule (CurrentAgent, FixedRule.Factory, outputContourObstacle, contourObstacleSupportCalculator);
			
			/*
			ruleAvoidCollisionWall.Parameters.PARTIAL_MATCH_ON = true;
			ruleAvoidCollisionWall.Parameters.DEFAULT_PARTIAL_MATCH_ON_THRESHOLD = 0.1;
			ruleAvoidCollisionWall.Parameters.PARTIAL_MATCH_THRESHOLD = 0.1;
			ruleAvoidCollisionWall.Parameters.POSITIVE_MATCH_THRESHOLD = 0.1;
			*/
			
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit (ruleContourObstacle);

            // Create deliver leaflet rule
			
            SupportCalculator deliverLeafletSupportCalculator = fixedRuleDelegateToDeliverLeaflet;
            FixedRule ruleDeliverLeaflet = AgentInitializer.InitializeActionRule (CurrentAgent, FixedRule.Factory, outputDeliverLeaflet, deliverLeafletSupportCalculator);		
			
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit (ruleDeliverLeaflet);
			
			
			
            // Disable Rule Refinement
            // CurrentAgent.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selection type will be probabilistic
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;

            // Define Probabilistic values
            CurrentAgent.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE  = 1;
            CurrentAgent.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE  = 0;
            CurrentAgent.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
			
			// Create goals for leaflet fetching process
			
			eatFood = World.NewGoalChunk(EAT_FOOD);			
			fulfillLeaflet = World.NewGoalChunk(FULFILL_LEAFLET);

			// Create the action chunks for setting each of those goals
			
            GoalStructureUpdateActionChunk setEatFood = World.NewGoalStructureUpdateActionChunk();
            GoalStructureUpdateActionChunk setFulfillLeaflet = World.NewGoalStructureUpdateActionChunk();

            setEatFood.Add (GoalStructure.RecognizedActions.SET_RESET, eatFood);
            setFulfillLeaflet.Add (GoalStructure.RecognizedActions.SET_RESET, fulfillLeaflet);
			
			// Create the drives
			
			DeficitChangeProcessor eatFoodAndFulfillLeafletDrivesDeficitChangeProcessor = EatFoodAndFulFillLeafletDrives_DeficitChange;

			// Eat Food drive
			
            FoodDrive eatFoodDrive = AgentInitializer.InitializeDrive (CurrentAgent, 
			                                                           FoodDrive.Factory, 
			                                                           .5, 
			                                                           eatFoodAndFulfillLeafletDrivesDeficitChangeProcessor);

            DriveEquation eatFoodDrive_equation = AgentInitializer.InitializeDriveComponent (eatFoodDrive, DriveEquation.Factory);

            eatFoodDrive.Commit (eatFoodDrive_equation);
			
			
			// Fulfill leaflet drive
			
            RecognitionAchievementDrive fulfillLeafletDrive = AgentInitializer.InitializeDrive (CurrentAgent,
			                                                                                    RecognitionAchievementDrive.Factory, 
			                                                                                    .5, 
			                                                                                    eatFoodAndFulfillLeafletDrivesDeficitChangeProcessor);

            DriveEquation fulfillLeafletDrive_equation = AgentInitializer.InitializeDriveComponent (fulfillLeafletDrive, DriveEquation.Factory);

            fulfillLeafletDrive.Commit(fulfillLeafletDrive_equation);
			
						
            CurrentAgent.Commit (eatFoodDrive);
			CurrentAgent.Commit (fulfillLeafletDrive);
			
			
			// Initialize the meta-cognitive module
			
			
			GoalSelectionModule gsm = AgentInitializer.InitializeMetaCognitiveModule (CurrentAgent, GoalSelectionModule.Factory);

            GoalSelectionEquation gse = AgentInitializer.InitializeMetaCognitiveDecisionNetwork (gsm, GoalSelectionEquation.Factory);

            gse.Input.Add (eatFoodDrive.GetDriveStrength());
            gse.Input.Add (fulfillLeafletDrive.GetDriveStrength());

            gse.Output.Add (setEatFood);
            gse.Output.Add (setFulfillLeaflet);

            gse.SetRelevance (setEatFood, eatFoodDrive, 1);	
            gse.SetRelevance (setFulfillLeaflet, fulfillLeafletDrive, 1);

            gsm.Commit (gse);

            CurrentAgent.Commit (gsm);

			CurrentAgent.MS.Parameters.CURRENT_GOAL_ACTIVATION_OPTION = MotivationalSubsystem.CurrentGoalActivationOptions.FULL;
			
			
			// Setup the movement network
			
			movementNet = AgentInitializer.InitializeImplicitDecisionNetwork (CurrentAgent, 
			                                                                  SimplifiedQBPNetwork.Factory, 
			                                                                  (EligibilityChecker) CustomEligibilityCheckerMovementNetwork);
			

			// Add the goals as input for the movement network
			
			movementNet.Input.Add (eatFood);
			movementNet.Input.Add (fulfillLeaflet);
			
			movementNet.Input.Add (InputWallAhead);
			movementNet.Input.Add (leafletCompletion);
			
			
			// Add the elements representing the current creature position
			
			movementNet.Input.Add (pos_x);
			movementNet.Input.Add (pos_y);
			
			
			// Add the elements representing the missing jewels in the existing leaflets
			
			movementNet.Input.Add (missing_red);
			movementNet.Input.Add (missing_green);
			movementNet.Input.Add (missing_blue);
			movementNet.Input.Add (missing_yellow);
			movementNet.Input.Add (missing_magenta);
			movementNet.Input.Add (missing_white);
			
			
			// Add the rule for rotating clockwise
			
			movementNet.Output.Add (OutputRotateClockwise);
			
			movementNet.Parameters.LEARNING_RATE = 1;
			
			CurrentAgent.Commit (movementNet);
			
            RefineableActionRule.GlobalParameters.SPECIALIZATION_THRESHOLD_1 = -.6;
            RefineableActionRule.GlobalParameters.GENERALIZATION_THRESHOLD_1 = -.1;
            RefineableActionRule.GlobalParameters.INFORMATION_GAIN_OPTION = RefineableActionRule.IGOptions.PERFECT;

        }
		
		
		
		
		/***
		 * 
		 * 
		 * SENSORY INFORMATION
		 * 
		 * 
		 */
		
        /// <summary>
        /// Make the agent perception. In other words, translate the information that came from sensors to a new type that the agent can understand
        /// </summary>
        /// <param name="sensorialInformation">The information that came from server</param>
        /// <returns>The perceived information</returns>
        /// 
        
        private SensoryInformation MakePerceptionFromSensorialInput(IList<Thing> sensorialInformation)
        {
			// Check if hasn't created yet the rule for going until the delivery spot
			
			if (deliverySpot == null)
			{
				deliverySpot = sensorialInformation.First (item => (item.CategoryId == Thing.CATEGORY_DeliverySPOT));
				deliverySpot.X1 = 30.0;
				deliverySpot.Y1 = 30.0;
				
				// Create an external action to go until the delivery spot
				
				ExternalActionChunk outputGoDeliverySpot = World.NewExternalActionChunk ("go_deliverySpot");
				
				outputGoDeliverySpot.Add (World.NewDimensionValuePair ("x", deliverySpot.X1));
				outputGoDeliverySpot.Add (World.NewDimensionValuePair ("y", deliverySpot.Y1));
				outputGoDeliverySpot.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.GO_POSITION.ToString()));
				
				// Update the movement network
			
				CurrentAgent.Retract (movementNet, Agent.InternalContainers.IMPLICIT_DECISION_NETWORKS);
				
				movementNet.Output.Add (outputGoDeliverySpot);
				
				CurrentAgent.Commit (movementNet);
				
				Console.WriteLine (">>>> Added delivery spot ({0}, {1}) into the agent", deliverySpot.X1, deliverySpot.Y1);
			}
			
			
			
			// Save the current creature state information
			currentCreature = (Creature) sensorialInformation.First (item => (item is Creature));

            // New sensory information
            SensoryInformation si = World.NewSensoryInformation(CurrentAgent);

            // Detect if we have a wall ahead
            Boolean wallAhead = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= CONTOUR_DISTANCE)).Any();
            
			double wallAheadActivationValue = wallAhead ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;
            
			creatureNear = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_CREATURE && 
			                                                    item.Name != currentCreature.Name &&
			                                                    item.DistanceToCreature <= TALK_DISTANCE)).Any();
							
			
			if (creatureNear == false)
			{
				communicationDone = false;
				currentCycle      = communicationCycles.OPEN_GREETING;
				
				communicationAttempts = 0;
				listeningAttempts     = 0;
				listenedWrongMessage  = 0;
			}
			
			
			// Stores the new jewels of foods for further reference

			
			
			foreach (Thing visualizedItem in sensorialInformation.Where (item => ((item.CategoryId == Thing.CATEGORY_FOOD) ||
			                                                                      (item.CategoryId == Thing.CATEGORY_NPFOOD) ||
			                                                                      (item.CategoryId == Thing.CATEGORY_PFOOD))))
			{
				Thing foundFood = foods.Find (item => item.Name == visualizedItem.Name);
				
				if (foundFood == null)
				{
					visualizedItem.lastSeen = CurrentCognitiveCycle;
					
					foods.Add (visualizedItem);
					
					CreateRulesForFood (visualizedItem);
				}
				else
				{
					foundFood.lastSeen = CurrentCognitiveCycle;
					
					// Check if food has been moved
					
					if ((foundFood.X1 != visualizedItem.X1) || (foundFood.Y1 != visualizedItem.Y1))
					{
						// Remove all the info related to the old food position...
						
						removeCollectedOrEaten (visualizedItem.Name);
						
						Thread.Sleep (100);
						
						
						// Now add info related to the new position
						
						foods.Add (visualizedItem);
						
						CreateRulesForFood (visualizedItem);
					}
				}
			}
			

			
			foreach (Thing visualizedItem in sensorialInformation.Where (item => item.CategoryId == Thing.CATEGORY_JEWEL))
			{
				Thing foundJewel = jewels.Find (item => item.Name == visualizedItem.Name);
				
				if (foundJewel == null)
				{
					visualizedItem.lastSeen = CurrentCognitiveCycle;
					
					jewels.Add (visualizedItem);
					
					CreateRulesForJewel (visualizedItem);					
				}
				else
				{
					foundJewel.lastSeen = CurrentCognitiveCycle;
					
					// Check if jewel has been moved
					
					if ((foundJewel.X1 != visualizedItem.X1) || (foundJewel.Y1 != visualizedItem.Y1))
					{
						// Remove all the info related to the old jewel position...
						
						removeCollectedOrEaten (visualizedItem.Name);
						
						Thread.Sleep (100);
						
						
						// Now add info related to the new position
						
						jewels.Add (visualizedItem);
						
						CreateRulesForJewel (visualizedItem);
					}
					else
					{						
						// Check if jewel has changed color
						
						if (foundJewel.Material.Color != visualizedItem.Material.Color)
						{
							// Update it
							
							foundJewel.Material.Color = visualizedItem.Material.Color;
						}
					}
				}
			}
			
			
			
			// Stores the received leaflets; updates the list just to make sure the most recent ones are stored
			
			leaflets = new List<Leaflet>();
			
			foreach (Thing visualizedItem in sensorialInformation.Where (item => item.CategoryId == Thing.CATEGORY_LEAFLET))
			{
				Leaflet newLeaflet = (Leaflet) visualizedItem;
				
				if (leaflets.Find (item => item.ID == newLeaflet.ID) == null)
				{
					leaflets.Add (newLeaflet);
				}
			}
			
			
						
			// Updates the leaflets counters
			
			completedLeaflet = CheckCompleteLeaflet ();
			
			
			
			
			//
			// Talking related activations
			//
				
			
			
			
			foreach (Thing visualizedItem in sensorialInformation.Where (item => item.CategoryId == Thing.CATEGORY_CREATURE))
			{
				// Check if creature is not yet in the hash...
				
				if (creatures.ContainsKey(visualizedItem.Name) == false)
				{
					// ...if not, add it.
					
					DimensionValuePair newDV = World.NewDimensionValuePair("creature", visualizedItem.Name);
					
					creatures.Add(visualizedItem.Name, newDV);
					
					// Update the talking network
			
					CurrentAgent.Retract (talkingNet, Agent.InternalContainers.IMPLICIT_DECISION_NETWORKS);
					
					talkingNet.Input.Add (newDV);
					
					CurrentAgent.Commit (talkingNet);
				}
					
				// Now, update the creature activation
				
				si.Add((DimensionValuePair) creatures[visualizedItem.Name], visualizedItem.DistanceToCreature / MAXIMUM_DISTANCE);
			}

			
			si.Add (hasTalkedBefore, justTalked == true ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION);
			si.Add (hasListenedBefore, justListened == true ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION);
			si.Add (hasListenedNoise, justListenedNoise == true ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION);
			si.Add (hasListenedNothing, justListenedNothing == true ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION);
			
			si.Add (commOpenGreeting, currentCycle == communicationCycles.OPEN_GREETING ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION);
			si.Add (commMessage, currentCycle == communicationCycles.MESSAGE ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION);
			si.Add (commCloseGreeting, currentCycle == communicationCycles.CLOSE_GREETING ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION);
			
			si.Add (InputWallAhead, wallAheadActivationValue);
			
			si.Add (leafletCompletion, completedLeaflet == null ? CurrentAgent.Parameters.MIN_ACTIVATION : CurrentAgent.Parameters.MAX_ACTIVATION);
			
			si.Add (pos_x, currentCreature.X1 / 800);
			si.Add (pos_y, currentCreature.Y1 / 600);
			        
			si.Add (missing_red, jewelsActivation[(int) jewelColors.red]);
			si.Add (missing_green, jewelsActivation[(int) jewelColors.green]);
			si.Add (missing_blue, jewelsActivation[(int) jewelColors.blue]);
			si.Add (missing_yellow, jewelsActivation[(int) jewelColors.yellow]);
			si.Add (missing_magenta, jewelsActivation[(int) jewelColors.magenta]);
			si.Add (missing_white, jewelsActivation[(int) jewelColors.white]);
				
			
			// Sets the drive stimulus, based on the current creature fuel level

			si[typeof(FoodDrive), FoodDrive.MetaInfoReservations.STIMULUS] = (MINIMUM_FUEL / currentCreature.Fuel) < 1 ? 0 : 1;
			si[typeof(RecognitionAchievementDrive), RecognitionAchievementDrive.MetaInfoReservations.STIMULUS] = ((currentCreature.Fuel - MINIMUM_FUEL) / MINIMUM_FUEL) > 0 ? 1 : 0;
			
            return si;
        }
		
		
		
		private Leaflet CheckCompleteLeaflet ()
		{
			Leaflet completeLeaflet = null;
			
			for (int i = 0; i < jewelsNeeded.Length; i++)
			{
				jewelsNeeded[i] = 0;
			}
			
			
			// Go through all the leaflets to find the first completed one
			
			foreach (Leaflet currentLeaflet in leaflets)
			{
				bool complete = true;
				
				foreach (LeafletItem jewelColor in currentLeaflet.items)
				{
					jewelsNeeded[(int) Enum.Parse(typeof (jewelColors), jewelColor.color, true)] += jewelColor.desiredQuantity; 
					
					if (jewelColor.desiredQuantity > collectedJewels.FindAll (item => item.Material.Color == jewelColor.color).Count)
					{
						complete = false;
					}
				}
				
				// Check if current leaflet is complete...
				
				if ((complete == true) && (completeLeaflet == null))
				{
					// ...if so, save it to return.
					
					completeLeaflet = currentLeaflet;
					
					Console.WriteLine ("Leaflet {0} is now completed !", currentLeaflet.ID);
				}
			}
			
			
			for (int i = 0; i < jewelsActivation.Length; i++)
			{
				int currentJewelsCount = collectedJewels.FindAll (item => item.Material.Color == Enum.GetName (typeof (jewelColors), i)).Count;
				
				if (jewelsNeeded[i] == 0 || jewelsNeeded[i] == currentJewelsCount)
				{                                               
					jewelsActivation[i] = 0.0;
				}
				else
				{
					jewelsActivation[i] = 1.0 - (currentJewelsCount + 0.1 / jewelsNeeded[i]);
				}
			}
			
			return completeLeaflet;
		}
		
		
		
		private void CreateRulesForFood (Thing whichFood)
		{
			// Common information
			
			DimensionValuePair foodName = World.NewDimensionValuePair ("food", whichFood.Name);
			DimensionValuePair foodX    = World.NewDimensionValuePair ("x", whichFood.X1);
			DimensionValuePair foodY    = World.NewDimensionValuePair ("y", whichFood.Y1);

			
			// External action for going until this food
			
			ExternalActionChunk outputGoFood = World.NewExternalActionChunk ("go_" + whichFood.Name);
			
			outputGoFood.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.GO_POSITION.ToString()));
			outputGoFood.Add (foodName);
			outputGoFood.Add (foodX);
			outputGoFood.Add (foodY);

			
			// Fixed Rule for eating this food
			
			ExternalActionChunk outputEatFood = World.NewExternalActionChunk ("eat_" + whichFood.Name);
			
			outputEatFood.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.EAT_FOOD.ToString()));
			outputEatFood.Add (foodName);
			outputEatFood.Add (foodX);
			outputEatFood.Add (foodY);

			SupportCalculator collectJewelSupportCalculator = FixedRuleDelegateToCollectOrEat;
			
            FixedRule ruleEatJewel = AgentInitializer.InitializeActionRule (CurrentAgent, 
			                                                                FixedRule.Factory, 
			                                                                outputEatFood, 
			                                                                collectJewelSupportCalculator);
			                                                                    			
			
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit (ruleEatJewel);
			
			
			// Update the movement network
			
			CurrentAgent.Retract (movementNet, Agent.InternalContainers.IMPLICIT_DECISION_NETWORKS);
			
			movementNet.Output.Add (outputGoFood);
			
			CurrentAgent.Commit (movementNet);
			
			Console.WriteLine (">>>> Added food {0} into the agent", whichFood.Name);

		}
		
		
		
		private void CreateRulesForJewel (Thing whichJewel)
		{
			// Common information
			
			DimensionValuePair jewelName = World.NewDimensionValuePair ("jewel", whichJewel.Name);
			DimensionValuePair jewelX    = World.NewDimensionValuePair ("x", whichJewel.X1);
			DimensionValuePair jewelY    = World.NewDimensionValuePair ("y", whichJewel.Y1);
			
			
			// External action for going until this jewel
			
			
			ExternalActionChunk outputGoJewel = (ExternalActionChunk) World.NewExternalActionChunk ("go_" + whichJewel.Name);
			
			
			outputGoJewel.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.GO_POSITION.ToString()));
			outputGoJewel.Add (jewelName);
            outputGoJewel.Add (World.NewDimensionValuePair ("color", whichJewel.Material.Color));
			outputGoJewel.Add (jewelX);
			outputGoJewel.Add (jewelY);
			

			// Fixed Rule for collecting this jewel
			
			ExternalActionChunk outputCollectJewel = World.NewExternalActionChunk ("collect_" + whichJewel.Name);
			
			outputCollectJewel.Add (World.NewDimensionValuePair ("type", ClarionAgentActionType.COLLECT_JEWEL.ToString()));
			outputCollectJewel.Add (jewelName);
			outputCollectJewel.Add (jewelX);
			outputCollectJewel.Add (jewelY);

			SupportCalculator eatFoddSupportCalculator = FixedRuleDelegateToCollectOrEat;
			
            FixedRule ruleCollectJewel = AgentInitializer.InitializeActionRule (CurrentAgent, 
			                                                                    FixedRule.Factory, 
			                                                                    outputCollectJewel, 
			                                                                    eatFoddSupportCalculator);
			                                                                    			
			
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleCollectJewel);
			
			
			// Update the movement network
			
			CurrentAgent.Retract (movementNet, Agent.InternalContainers.IMPLICIT_DECISION_NETWORKS);
			
			movementNet.Output.Add (outputGoJewel);
			
			CurrentAgent.Commit (movementNet);
			
			Console.WriteLine (">>>> Added jewel {0} into the agent", whichJewel.Name);
		}
		
		
		

		
		private void removeCollectedOrEaten (string whichThing)
		{	
			string token;
			bool isJewel = false;
			
			// Remove thing related rules
			
			IEnumerable<ICommitable> allRules = CurrentAgent.GetInternals (Agent.InternalContainers.ACTION_RULES);
						
			if (whichThing.IndexOf ("Jewel") >= 0)
			{
				token   = "collect_";
				isJewel = true;
			}
			else
			{
				token = "eat_";
			}
			
			
			foreach (ICommitable rule in allRules)
			{
				if ((rule is FixedRule) && ((FixedRule) rule).OutputChunk.LabelAsIComparable.ToString() == token + whichThing)
				{
					CurrentAgent.Retract (rule, Agent.InternalContainers.ACTION_RULES);
					CurrentAgent.Remove (rule);
				}
			}
	
			
			// Remove output from movement network
			
			CurrentAgent.Retract (movementNet, Agent.InternalContainers.IMPLICIT_DECISION_NETWORKS);
			
			
			ExternalActionChunk outputGoThing = (ExternalActionChunk) World.GetActionChunk ("go_" + whichThing);
			
			movementNet.Output.Remove (outputGoThing);
			
			CurrentAgent.Commit (movementNet);			
			
			
			IEnumerator<DimensionValuePair> iItems = outputGoThing.GetEnumerator();
			
			while (iItems.MoveNext())
			{
				DimensionValuePair item = iItems.Current;
				
				World.Remove (item);
			}
				
			World.Remove (outputGoThing);
			
				
			ExternalActionChunk outputCollectOrEat = (ExternalActionChunk) World.GetActionChunk (token + whichThing);
			
			
			iItems = outputCollectOrEat.GetEnumerator();
			
			while (iItems.MoveNext())
			{
				DimensionValuePair item = iItems.Current;
				
				World.Remove (item);
			}
									
			World.Remove (outputCollectOrEat);
			
			
			
			if (isJewel == true)
			{
				Thing whichThingToRemove = jewels.Find (item => item.Name == whichThing);
	
				collectedJewels.Add (whichThingToRemove);
				
				jewels.Remove (whichThingToRemove);
			}
			else
			{
				Thing whichThingToRemove = foods.Find (item => item.Name == whichThing);

				foods.Remove (whichThingToRemove);
			}
			
			
			Console.WriteLine (">>>>>>>>>>>>>>>>>>>> Removed thing {0}", whichThing);
		}

		
		
		
		
        #endregion

        #region Delegate Methods
		
		#region Deficit change 
		
		// Update the drive defcit according to the current goal set
		
		private double EatFoodAndFulFillLeafletDrives_DeficitChange (ActivationCollection currentInput, Drive target)
		{
			GoalChunk currentGoal = CurrentAgent.CurrentGoal;
			
            if (currentGoal != null)
            {
				// Check if the current goal is aligned with the drive which deficit is being updated...
				
                if ((currentGoal.LabelAsIComparable.ToString() == FULFILL_LEAFLET && target is RecognitionAchievementDrive) ||
                    (currentGoal.LabelAsIComparable.ToString() == EAT_FOOD && target is FoodDrive))
				{
					
					// ...if so, decrease the deficit by a 0.1% rate
					
                    target.Parameters.DEFICIT_CHANGE_RATE = .999;
				}
                else
				{	
					// ...if not, increase the drive deficit by a 0.1% rate
					
                    target.Parameters.DEFICIT_CHANGE_RATE = 1.001;
				}
            }

            return target.Deficit * target.Parameters.DEFICIT_CHANGE_RATE;
		}
		
		#endregion
			
		
        #region Fixed Rules	
		
		
		
		private double FixedRuleDelegateToContourObstacle (ActivationCollection currentInput, Rule target)
        {
			double activation = ((currentInput.Contains (InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0;
			
            // See partial match threshold to verify what are the rules available for action selection
			
			Console.WriteLine ("Contour: {0}", activation);
			
            return activation;
        }

		
		
		
		private double fixedRuleDelegateToDeliverLeaflet (ActivationCollection currentInput, Rule target)
        {
			Double distance = Double.PositiveInfinity;
			double activation = 0.0;
			
			distance = Math.Sqrt(Math.Pow (currentCreature.X1 - deliverySpot.X1, 2) + Math.Pow (currentCreature.Y1 - deliverySpot.Y1, 2)); 
				
			activation = ((distance < DELIVER_DISTANCE) && (completedLeaflet != null)) ? 1.0 : 0.0;
			
			
			//Console.WriteLine ("DELIVER LEAFLET - Activation: {0}", activation);
			
            return activation;
        }

		
		
		private double FixedRuleDelegateToCollectOrEat (ActivationCollection currentInput, Rule target)
        {
			Double distance = Double.PositiveInfinity;
			double activation = 0;
			
			ExternalActionChunk outputCollectOrEat = (ExternalActionChunk) target.OutputChunk;
			
			double thing_x = System.Convert.ToDouble (outputCollectOrEat.First (item => item.AsDimensionValuePair.Dimension.ToString() == "x").Value.AsIComparable);
			double thing_y = System.Convert.ToDouble (outputCollectOrEat.First (item => item.AsDimensionValuePair.Dimension.ToString() == "y").Value.AsIComparable);

			// Get the thing name, removing the leading "go_"...
			
			string name = outputCollectOrEat.LabelAsIComparable.ToString().Substring (outputCollectOrEat.LabelAsIComparable.ToString().IndexOf ("_") + 1);
				
			
			if (((name.IndexOf ("Jewel") >= 0) && (jewels.Find (item => item.Name == name) != null)) ||
			    ((name.IndexOf ("Food") >= 0) && (foods.Find (item => item.Name == name) != null)))
			{
				distance = Math.Sqrt(Math.Pow (thing_x - currentCreature.X1, 2) + Math.Pow (thing_y - currentCreature.Y1, 2)); 
				
				activation = distance < COLLECT_DISTANCE ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;
			}
			
			//Console.WriteLine ("COLLECT/EAT ({0}, {1}) - thing: {2} - distance: {3} - Activation: {4}", thing_x, thing_y, name, distance, activation);
			
            return activation;
        }

		
		#endregion
		
		
		#region Eligibility checker
		
		private bool CustomEligibilityCheckerCommNetwork (ActivationCollection currentInput, ClarionComponent target)
		{			
			Console.WriteLine("\n\n\n>>>>>>>>>>>>>>>>>CommNetwork - creatureNear: " + creatureNear + "  communicatonDone: " + communicationDone);
			
			
			return (creatureNear == true) && (communicationDone == false);
		}

		
		private bool CustomEligibilityCheckerMovementNetwork (ActivationCollection currentInput, ClarionComponent target)
		{		
			Console.WriteLine("\n\n\n>>>>>>>>>>>>>>>>>MovementNetwork - creatureNear: " + creatureNear + "  communicatonDone: " + communicationDone);
			
			
			return (creatureNear == false) || (communicationDone == true);
		}

		
		
		
		#endregion
		
		
		
        #endregion

		
		
		
        #region Run Thread Method

        private void RunThread(object obj)
        {
			ExternalActionChunk chosen = ExternalActionChunk.DO_NOTHING;
			
			double lastMovementFeedback = 0.0;
			
			bool tryingToContour = false;
			
			Thing lastTargetPosition = null;
			
		
			
			Console.WriteLine("Starting Cognitive Cycle ... press CTRL-C to finish !");
            // Cognitive Cycle starts here getting sensorial information
			
            while (CurrentCognitiveCycle != MaxNumberOfCognitiveCycles)
            {   
				//Console.WriteLine ("+++ Cicle: {0}", CurrentCognitiveCycle);
				
				
				if (Console.KeyAvailable == true)
				{
					if (Console.ReadKey (true).Key == ConsoleKey.Spacebar)
					{
						OnNewExternalActionSelected (ClarionAgentActionType.CREATE_THINGS, null, null);
					}
				}
				
                // Get current sensorial information                    
                IList<Thing> sensorialInformation = OnNewVisualSensorialInformation();

                // Make the perception
                SensoryInformation si = MakePerceptionFromSensorialInput(sensorialInformation);
				
				
                //Perceive the sensory information
                CurrentAgent.Perceive(si);

                //Choose an action
				
				
            	chosen = CurrentAgent.GetChosenExternalAction(si);
				
				if (chosen != ExternalActionChunk.DO_NOTHING)
				{
	                // Get the selected action
	                String actionLabel = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "type").Value.AsIComparable;
	
					
					Console.WriteLine ("\n\nHas choosen {0} action. Label: {1}\n\n", actionLabel, chosen.LabelAsIComparable.ToString());
					
					
					GoalChunk currentGoal = CurrentAgent.CurrentGoal;
					
					//Console.WriteLine ("Current goal: {0}", currentGoal);
	
					
					
					ClarionAgentActionType actionType = (ClarionAgentActionType) Enum.Parse(typeof(ClarionAgentActionType), actionLabel, true);

	
	                // Call the output event handler
	                if (OnNewExternalActionSelected != null)
	                {
						switch (actionType)
						{
						case ClarionAgentActionType.GO_POSITION:
							
						    Thing targetPosition = null;
						
							tryingToContour = false;
													
							double x = (double) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "x").Value.AsIComparable;
					        double y = (double) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "y").Value.AsIComparable;
							
							Console.WriteLine ("({0}, {1})", x, y);
						

						    if (chosen.LabelAsIComparable.ToString() == "go_deliverySpot")
							{
								if (completedLeaflet != null)
								{
									CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MAX_ACTIVATION);
								
									lastMovementFeedback = 1.0;
								
								    targetPosition = deliverySpot;
								
									Console.WriteLine (". Going to delivery spot...");
								}
								else
								{
									CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MIN_ACTIVATION);
								
									Console.WriteLine (". Ignoring request to go to delivery spot, since there is no completed leaflet...");
								}
							}
							else if ((chosen.LabelAsIComparable.ToString()).IndexOf ("Jewel") >= 0)
							{		
								if (currentGoal != null ? (currentGoal.LabelAsIComparable.ToString() == FULFILL_LEAFLET) : true)
								{								
									string jewelColor = (string) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "color").Value.AsIComparable;
															
									int index = (int) Enum.Parse(typeof(jewelColors), jewelColor, true);
								
									double needValue = jewelsActivation[index] * 0.5;
									double distance  = 1.0 - (Math.Sqrt (Math.Pow (x - currentCreature.X1, 2) + Math.Pow (y - currentCreature.Y1, 2)) / MAXIMUM_DISTANCE);
								
									double currentFeedbackValue = (needValue + distance) / 2;
									
									Console.WriteLine (". Jewel - lastMovementFeedback: {0}, currentFeedbackValue: {1}", lastMovementFeedback, currentFeedbackValue);
								
								
									if (currentFeedbackValue > lastMovementFeedback)
									{								
										String jewelName = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "jewel").Value.AsIComparable;
									
										targetPosition = jewels.Find (item => item.Name == jewelName);
									
									
										if ((targetPosition == lastTargetPosition) && (targetPosition.lastSeen < CurrentCognitiveCycle - CYCLES_TOLERANCE))
										{
											// The jewel has been taken by another creature...
										
										    Console.WriteLine (". Jewel {0} - lastSeen: {1} - current: {2}. Already eaten by other creature...", 
										                       targetPosition.Name, targetPosition.lastSeen, CurrentCognitiveCycle);

											CurrentAgent.ReceiveFeedback (si, 0.0);
										
											Thread.Sleep(100);

											removeCollectedOrEaten (targetPosition.Name);
										
											targetPosition = null;
										
											lastMovementFeedback = 0.0;
										}
										else
										{
											CurrentAgent.ReceiveFeedback (si, currentFeedbackValue);
										
											// Wait for the feedback before crushing the SI...
											
											Thread.Sleep(100);
											
										    lastMovementFeedback = currentFeedbackValue;										
										}
									}
									else
									{
										// Give a bad feedback...
									
										CurrentAgent.ReceiveFeedback (si, 0.0);
									
									    Console.WriteLine (". Jewel - Bad movement choice. Ignoring it...");
									}
									
								
									Console.WriteLine (". Jewel - needValue: {0}, distance: {1}, feedback: {2}", needValue, distance, currentFeedbackValue);
								}
								else
								{
									// Give bad feedback, since should be going after food...
								
									CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MIN_ACTIVATION);
								
								    Console.WriteLine (". Jewel - Bad movement choice; should be going after food. Ignoring it...");
								}
							}
							else if ((chosen.LabelAsIComparable.ToString()).IndexOf ("Food") >= 0)
							{		
								if ((currentGoal != null ? (currentGoal.LabelAsIComparable.ToString() == EAT_FOOD) : false))
								{								
									double distance  = 1.0 - (Math.Sqrt (Math.Pow (x - currentCreature.X1, 2) + Math.Pow (y - currentCreature.Y1, 2)) / MAXIMUM_DISTANCE);
								
									double currentFeedbackValue = distance;
									
									Console.WriteLine (". Food - lastMovementFeedback: {0}, currentFeedbackValue: {1}", lastMovementFeedback, currentFeedbackValue);
								
								
									if (currentFeedbackValue > lastMovementFeedback)
									{								
										String foodName = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "food").Value.AsIComparable;
									
										targetPosition = foods.Find (item => item.Name == foodName);
									
										if ((targetPosition == lastTargetPosition) && (targetPosition.lastSeen < CurrentCognitiveCycle - CYCLES_TOLERANCE))
										{
											// The food has been eaten by another creature...
										
										    Console.WriteLine (". Food {0} - lastSeen: {1} - current: {2}. Already eaten by other creature...", 
										                       targetPosition.Name, targetPosition.lastSeen, CurrentCognitiveCycle);

											CurrentAgent.ReceiveFeedback (si, 0.0);
										
											Thread.Sleep(100);
										
											removeCollectedOrEaten (targetPosition.Name);
										
											targetPosition = null;
										
											lastMovementFeedback = 0.0;
										}
										else
										{
											CurrentAgent.ReceiveFeedback (si, currentFeedbackValue);
										
											// Wait for the feedback before crushing the SI...
											
											Thread.Sleep(100);
											
										    lastMovementFeedback = currentFeedbackValue;
										}										
									}
									else
									{
										// Give a bad feedback...
									
										CurrentAgent.ReceiveFeedback (si, 0.0);
									
									    Console.WriteLine (". Food - Bad movement choice. Ignoring it...");
									}
									
								
									Console.WriteLine (". Food - distance: {0}, feedback: {1}", distance, currentFeedbackValue);
								}
								else
								{
									// Give bad feedback, since should be going after jewels...
								
									CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MIN_ACTIVATION);
								
								    Console.WriteLine (". Food - Bad movement choice; should be going after jewels. Ignoring it...");
								}
							}
						
						
							if (targetPosition != null)
							{
								OnNewExternalActionSelected (actionType, targetPosition, null);
							
								lastTargetPosition = targetPosition;
							}
						
							break;
						
						
						case ClarionAgentActionType.COLLECT_JEWEL:
							
							tryingToContour = false;
						
							String jewelName = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "jewel").Value.AsIComparable;
						
							Thing whichJewel = jewels.Find (item => item.Name == jewelName);
						
							if (whichJewel != null)
							{
								OnNewExternalActionSelected (actionType, whichJewel, null);

								removeCollectedOrEaten (jewelName);
							
								// Reset movement feedback controller
							
								lastMovementFeedback = 0.0;								
							
							
								// Wait for a little bit to avoid WorldServer3D to still "see" the collected jewel...
							
								Thread.Sleep (100);
							}
							else
							{
								Console.WriteLine ("Target jewel is not found...");
							}
													
						
							break;
						
						
						
						case ClarionAgentActionType.EAT_FOOD:
						
							String foodName = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "food").Value.AsIComparable;
						
							Thing whichFood = foods.Find (item => item.Name == foodName);
						
							tryingToContour = false;
						
							if (whichFood != null)
							{
								OnNewExternalActionSelected (actionType, whichFood, null);

								removeCollectedOrEaten (foodName);
							
							
								// Reset movement feedback controller
							
								lastMovementFeedback = 0.0;			
							
								// Wait for a little bit to avoid WorldServer3D to still "see" the eaten food...

								Thread.Sleep (100);
							}
							else
							{
								Console.WriteLine ("Target food is not found...");
							}
													
						
							break;
						
						
						
						case ClarionAgentActionType.ROTATE_CLOCKWISE:
						
//								if ((si[InputWallAhead] == CurrentAgent.Parameters.MAX_ACTIVATION) || lastMovementFeedback == 0.0)
						
							tryingToContour = false;
						
							if (lastMovementFeedback == 0.0)
							{
	                    		OnNewExternalActionSelected (actionType, null, null);
						
								CurrentAgent.ReceiveFeedback (si, 1.0);
							
								lastMovementFeedback = 0.0;
							}
							else
							{
								// Give bad feedback...
							
								CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MIN_ACTIVATION);
							
								Console.WriteLine ("No reason to request rotate clockwise...");
							}
													
							break;
						
						
						case ClarionAgentActionType.CONTOUR_OBSTACLE:
						
							if (si[InputWallAhead] == CurrentAgent.Parameters.MAX_ACTIVATION)
							{
								IEnumerable<Thing> whichThings = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= CONTOUR_DISTANCE));
								Thing whichThing = ((whichThings.Count() > 1) || (tryingToContour == true)) ? null : whichThings.First ();
	                    		
								OnNewExternalActionSelected (actionType, whichThing, null);
						
								//CurrentAgent.ReceiveFeedback (si, 1.0);
							
								lastMovementFeedback = 0.0;
							
								// Let the creature move away from the obstacle
							
								//if (whichThing == null)
								//{	
								//	Thread.Sleep (100);
								//}
							
								tryingToContour = !tryingToContour;
							}
							else
							{
								// Give bad feedback...
							
								CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MIN_ACTIVATION);
							
								Console.WriteLine ("No obstacle to contour...");
							}
						
							break;
						
						
						case ClarionAgentActionType.DELIVER_LEAFLET:						
						
						
							tryingToContour = false;
						
                    		OnNewExternalActionSelected (actionType, completedLeaflet, null);

						
							// Now remove the corresponding leaflet jewels from the collectedJewels list
						
							foreach (LeafletItem jewelColor in completedLeaflet.items)
							{
								Console.WriteLine ("Removing {0} {1} jewels from collected", jewelColor.desiredQuantity, jewelColor.color);
							
								for (int i = 0; i < jewelColor.desiredQuantity; i++)
								{
									Thing toRemove = collectedJewels.First (item => item.Material.Color == jewelColor.color);
									collectedJewels.Remove (toRemove);
								}
							}
						
							
							// Zeroes the last feedback...
						
							lastMovementFeedback = 0.0;
							
							break;
							
							
							
						case ClarionAgentActionType.TALK:
							
							Console.WriteLine("\n\n[COMM] >>>>>>>>>> TALKING\n");
							
							String message = (String) chosen.First (item => item.AsDimensionValuePair.Dimension.ToString() == "message").Value.AsIComparable;

							if (justTalked == false)
							{	
								// Check if it is the correct message
								
								if ((message.Equals(communicationMessages[(int) communicationCycles.OPEN_GREETING]) && currentCycle == communicationCycles.OPEN_GREETING) ||
								    (message.Equals(communicationMessages[(int) communicationCycles.CLOSE_GREETING]) && currentCycle == communicationCycles.CLOSE_GREETING) ||
								    (message.Equals(communicationMessages[(int) communicationCycles.MESSAGE]) && currentCycle == communicationCycles.MESSAGE))
								{
									
									OnNewExternalActionSelected (actionType, null, message);
									
									justTalked = true;
									
									
									// Wait for the talking period
									
									Thread.Sleep(1000);
									
									
									if (justListened == true)
									{
										Console.WriteLine ("\n\n[COMM] TALKING CORRECT CYCLE: Talked < {0} >\n", message);
											
										if (currentCycle == communicationCycles.CLOSE_GREETING)
										{
											communicationDone = true;
											
											CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MAX_ACTIVATION);
											
											Console.WriteLine ("\n\n[COMM] END OF CONVERSATION\n");
										}
										else
										{
											CurrentAgent.ReceiveFeedback (si, 0.5);
											
											Console.WriteLine ("\n\n[COMM] COMPLETED CYCLE {0}\n", Enum.GetName (typeof (communicationCycles), currentCycle));											
											
											currentCycle++;
											
											// Reset everything when a cycle is completed
											
											justTalked = false;
										}
									}
									else
									{
										CurrentAgent.ReceiveFeedback (si, 0.5);
										
										Console.WriteLine ("\n\n[COMM] TALKING CORRECT CYCLE: Talked < {0} >\n", message);
									}
									
									justListened        = false;
									justListenedNoise   = false;
									justListenedNothing = false;									
								}
								else
								{
									// Trying to say wrong message. 
									// Give bad feedback...
								
									CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MIN_ACTIVATION);
																
									Console.WriteLine("\n\n[COMM] Trying to talk < {0} >, but should be talking < {1} >\n", 
									                  message, Enum.GetName (typeof (communicationCycles), currentCycle));
								}
								
								listeningAttempts    = 0;
								listenedWrongMessage = 0;
							}
							else
							{
								// Should be listening now...
								// Give bad feedback...
							
								CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MIN_ACTIVATION);
															
								Console.WriteLine("\n\n[COMM] Trying to talk < {0} >, but should be listening\n", message);
								
								justListened        = false;
								justListenedNoise   = false;
								justListenedNothing = false;								
							}
							
							communicationAttempts++;
							
							Console.WriteLine("\n\n[COMM] <<<<<<<<<< Talking. Total communication attempts: {0}\n", communicationAttempts);					
							
							break;
							
							
						case ClarionAgentActionType.LISTEN:
							
							Console.WriteLine("\n\n[COMM] >>>>>>>>>> LISTENING\n");
							
							if (justListened == false)
							{
								String response = OnNewExternalActionSelected (actionType, null, null);
								
                    			string[] tokens = response.Split(new char[] { ' ', '<', '>', ':' }, StringSplitOptions.RemoveEmptyEntries);
								
								if (tokens[0].Equals("Ok."))
								{
									justListened = true;
									
									// Check if it is the correct message
									
									if ((tokens[3].Equals(communicationMessages[(int) communicationCycles.OPEN_GREETING]) && currentCycle == communicationCycles.OPEN_GREETING) ||
									    (tokens[3].Equals(communicationMessages[(int) communicationCycles.CLOSE_GREETING]) && currentCycle == communicationCycles.CLOSE_GREETING) ||
									    (tokens[3].Equals(communicationMessages[(int) communicationCycles.MESSAGE]) && currentCycle == communicationCycles.MESSAGE))
									{
										if (justTalked == true) 
										{
											Console.WriteLine("\n\n[COMM] LISTENING CORRECT CYCLE: Creature {0} talked < {1} >\n", tokens[1], tokens[3]);
												
											if (currentCycle == communicationCycles.CLOSE_GREETING)
											{
												communicationDone = true;
												
												CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MAX_ACTIVATION);
										
												Console.WriteLine ("\n\n[COMM] END OF CONVERSATION\n");
											}
											else
											{
												CurrentAgent.ReceiveFeedback (si, 0.5);
	
												Console.WriteLine ("\n\n[COMM] COMPLETED CYCLE {0}\n", Enum.GetName (typeof (communicationCycles), currentCycle));											
											
												currentCycle++;
												
												// Reset everything when a cycle is complete
												
												justListened = false;
											}
										}
										else
										{
											CurrentAgent.ReceiveFeedback (si, 0.5);

											Console.WriteLine("\n\n[COMM] LISTENING CORRECT CYCLE: Creature {0} talked < {1} >\n", tokens[1], tokens[3]);
										}
										
										listenedWrongMessage = 0;
									}
									else
									{
										// Listened to wrong message.
										
										// Give neutral feedback, since it seemed correct to listen
										
										CurrentAgent.ReceiveFeedback (si, 0.5);
										
										Console.WriteLine("\n\n[COMM] WRONG CYCLE: Listened to < {0} > while should have listened < {1} >",
										                  tokens[3],
										                  Enum.GetName (typeof (communicationCycles), currentCycle));
										
										// Check if should wait for synchronization
										
										if (listenedWrongMessage >= WRONG_MESSAGE_THRESHOLD)
										{
											// Check if should return to the talking creature cycle
											
											int i;
											
											for (i = 0; i < (int) communicationCycles.NUM_OF_CYCLES && (communicationMessages[i].Equals(tokens[3]) == false); i++);
	
											if (i < (int) currentCycle)
											{
												// Return to the talking cycle
												
												currentCycle = (communicationCycles) i;
												
												Console.WriteLine ("\n\n[COMM] Returning to the talking cycle {0}\n", Enum.GetName (typeof (communicationCycles), currentCycle));
											}
											
											listenedWrongMessage = 0;
										}
										else
										{
											listenedWrongMessage++;
											
											Console.WriteLine("\n\n[COMM] Listening wrong message {0}\n", listenedWrongMessage);
										}
										
										
										
										// Wait for a little bit for the current message go away
										
										//Thread.Sleep(500);
										
										// Reset everything, since the communication is not synced
										
										justListened = false;
									}
									
									justListenedNoise   = false;
									justListenedNothing = false;
										
									justTalked = false;
									
									listeningAttempts = 0;
									
									
									// Wait for the listening period
									
									Thread.Sleep (500);
								}
								else
								{
									if (tokens[1].Equals("Talking"))
									{
										// Just ignore, keeping the current status
										
										Console.WriteLine("\n\n[COMM] Still in the talking time.\n");
									}
									else
									{
										if (tokens[1].Equals("Creatures"))
										{
											justListenedNoise   = true;
											justListenedNothing = false;
										
											
											// Give neutral feedback, since it seemed correct to listen
											
											CurrentAgent.ReceiveFeedback (si, 0.5);
											
											Console.WriteLine("\n\n[COMM] Multiple creatures talking ({0}).\n", tokens.Length - 7);
										}
										else
										{
											// Silence
											
											justListenedNothing = true;
											justListenedNoise   = false;
											
											if (justTalked == false)
											{
												// Give a little bellow neutral feedback since it could have talked instead.
												
												CurrentAgent.ReceiveFeedback (si, 0.2);
											}
											else
											{
												// Give neutral feedback, since it seemed correct to listen
												
												CurrentAgent.ReceiveFeedback (si, 0.5);
											}
											
											Console.WriteLine("\n\n[COMM] No one talked.\n");
										}
										
										if (listeningAttempts >= LISTENING_THRESHOLD)
										{											
											Console.WriteLine("\n\n[COMM] Listening attempts threshold reached {0}.\n", listeningAttempts);
											
											listeningAttempts = 0;
											justTalked        = false;
										}
										else
										{
											listeningAttempts++;
											
											Console.WriteLine("\n\n[COMM] Listening attempts {0}\n", listeningAttempts);
										}
									}
								}
							}	
							else
							{
								// Should be talking now...
								// Give bad feedback...
							
								CurrentAgent.ReceiveFeedback (si, CurrentAgent.Parameters.MIN_ACTIVATION);
															
								Console.WriteLine("\n\n[COMM] Trying to listen, but should be talking\n");
								
								justTalked = false;
							}
									
							
							communicationAttempts++;							

							Console.WriteLine("\n\n[COMM] <<<<<<<<<< Listening. Total communication attempts: {0}\n", communicationAttempts);					
							
							break;	
						}
	                }
				}
				else
				{
					Console.WriteLine ("Do nothing...");
				}
                			
                // Increment the number of cognitive cycles
                CurrentCognitiveCycle++;

                //Wait to the agent accomplish his job
                if (TimeBetweenCognitiveCycles > 0)
                {
                    Thread.Sleep(TimeBetweenCognitiveCycles);
                }
            }
        }

        #endregion

    }
}
